from thesplog.system.utilis import thesplog
import logging
import socket


class OutboundNotificationPoint(object):
    "Whom to notify when the transmit has reached the specified point"
    def __init__(self, index, notify):
        self._index  = index
        self._notify = notify
    def sentMore(self, amount):
        self._index = max(0, self._index - amount)
    def readyToNotify(self):
        return not self._index
    def doNotification(self):
        self._notify()



class SelectInfo(object):
    "Returned to the select logic to indicate the interesting elements for this socket stream"
    def __init__(self, fd, notifyOnSendable, notifyOnReadable, timeout):
        self.fd               = fd
        self.notifyOnSendable = notifyOnSendable
        self.notifyOnReadable = notifyOnReadable
        self.timeout          = timeout


class SelectResult(object):
    "Passed by core selector to indicate select response status for this socket stream"
    def __init__(self, canReceive=False,
                 canSend=False,
                 hasException=False,
                 timedOut=False):
        self.canReceive   = canReceive
        self.canSend      = canSend
        self.hasException = hasException
        self.timedOut     = timedOut



class SockMsgStream(object):
    def __init__(self, remoteAddr):
        self._sockstrm = SockStream(remoteAddr)
        self._intentNumNext = 0
        self._intentsBeingSent = {}
    def _nextTransmitStep(intent):  # only for intent.targetAddr == self.remoteAddr
        self._intentsBeingSent[self._intentNumNext] = intent
        self._sockstrm.addOutbound('{<{TX#%d:'%self._intentNumNext + intent.serMsg,
                                   itertools.partial(self._intentDataSent,
                                                     intentNum=self._intentNumNext))
        self._intentNumNext += 1
    def _intentDataSent(self, intentNum, sendStatus=SendStatus.Sent):
        del self._intentsBeingSent[intentNum]
        self._intentsAwaitingAck[intentNum] = self._intentsBeingSent[intentNum]
        intent.result = status
        intent.completionCallback()
    # KWQ: on timeout, reset self._sockstrm and retransmit all on _intentsBeingSent and _intentsAwaitingAck


class SockStream(object):
    def __init__(self, remoteAddr):
        self._outboundBuf = b''
        self._outboundPosition = 0
        self._outboundNotificationPoints = []
        self._inboundBuf = b''
        self._inboundNotify = None
        self._inboundExpectedSize = 20
        self._socketobj = RawSocket(remoteAddr)
        self._keepOpen = False
        self._remoteAddrs = [remoteAddr]


    def addOutbound(self, bufdata, notify):
        origlen = len(self._outboundBuf)
        self._outboundBuf += bufdata
        self._outboundNotificationPoints.append(OutboundNotificationPoint(origlen + len(bufdata), notify))
        self._initiateOutbound()


    def _initiateOutbound(self):
        self._reduceOutbound(self._socketobj.RSsend(None, self._outboundBuf))


    def _reduceOutbound(self, amountSent):
        if amountSent:
            for each in self._outboundNotificationPoints:
                each.sentMore(amountSent)
            self._outboundBuf = self._outboundBuf[amountSent:]
            if not self._outboundBuf:
                self._checkConn()
            done, pending = partition(lambda ONP: ONP.readyToNotify(),
                                      self._outboundNotificationPoints)
            self._outboundNotificationPoints = pending
            for each in done:
                each.doNotification(SendStatus.Sent)


    def registerInboundNotify(self, notify):
        self.inboundNotify = notify


    def keepOpen(self, newState=True):
        self._keepOpen = newState
        self._checkConn()

    def _checkConn(self):
        # Open or close the connection as needed
        if not self._keepOpen:
            if self._socketIsOpen():
                if not self._outboundBuf and not self._inboundBuf:
                    self._closeSocket()
        else:
            if not self._socketIsOpen():
                self._openSocket()

    def _socketIsOpen(self):
        return self._socketobj._socketIsOpen()

    def _openSocket(self):
        self._socketobj.RSopen()

    def _closeSocket(self):
        self._socketobj.RSclose()

    def selectOn(self):
        return self._socketobj.RSselectOn(self._outboundBuf,
                                          self._inboundExpectedSize)

    def select_work(self, selectResult):
        self._reduceOutbound(self._socketobj.RSsend(selectResult, self._outboundBuf))
        rcvd = self._socketobj.recv(selectResult, self._inboundExpectedSize or 20)
        if rcvd:
            self._addInbound(rcvd)
        # if hasException:
        #     self._closeSocket()

        # if hasIncoming and self._sockState == self.State.PendingConnect:
        #     self._sockState = self.State.Connected
        # ... lots of stuffz here.  Move this up to SockStream, or provide this will callbacks to same?




class RawSocket(object):
    class State(object):
        Closed = 0
        PendingConnect = 1
        Connected = 2


    def __init__(self, rmtAddr):
        self._rmtAddr = rmtAddr
        self._socket  = None
        self._sockState = self.State.Closed
        # KWQ: backoffPause here...?


    def _socketIsOpen(self):
        "Returns true if socket is not completely closed"
        return self._sockState == self.State.Closed


    def RSselectOn(self, haveTXData=False, haveRXData=False):
        if self._socket:
            return SelectInfo(haveRXData, haveTXData, None)  # KWQ: timeout?
        return SelectInfo(False, False, None)   # KWQ: timeout?


    def RSclose(self):
        # Try this even if the socket is thought to be closed at the moment
        try:
            self._socket.shutdown(socket.SHUT_RDWR)
        except socket.error as ex:
            if ex.errno != errno.ENOTCONN:
                # The close below should trump any error, but these should be
                # noted anyhow
                thesplog('Error during shutdown of socket %s: %s',
                         self._socket, ex, level=logging.WARNING)
        self._socket.close()
        self._sockState = self.State.Closed


    def RSopen(self):
        assert self._sockState == self.State.Closed
        self._sockState = self.State.PendingConnect
        self._socket = socket.socket(*self.rmtAddr.addressDetails.socketArgs)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._socket.setblocking(0)
        # Disable Nagle to transmit headers and acks asap; our sockets
        # are usually small command-and-control exchanges as opposed
        # to unidirectional streaming.
        self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        try:
            self._socket.connect(*self._rmtAddr.addressDetails.connectArgs)
        except socket.error as err:
            if not err_inprogress(err.errno):
                thesplog('Socket connect failure %s to %s on %s',
                         err, self._rmtAddr, self._socket,
                         level=logging.WARNING)
                # KWQ: errorCallback?
                self._sockState = self.State.Closed


    def RSsend(self, selectResult=None, self._outboundBuf):
        # nb. selectResult will be None when initiating a new transmit
        # that might require opening the socket.
        if not selectResult:
            if self._sockState == self.State.Closed:
                self.RSopen()
            return 0
        if selectResult.canSend:
            return self._socket.send(self._outboundBuf)
        if self._outboundBuf and self._sockState == self.State.Closed:
            self.RSopen()
        return 0

    def RSrecv(self,
